"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

from __future__ import annotations

import copy
import hashlib
import json
import re
from keyword import iskeyword
from typing import Dict, List, Optional, Union, Any, Iterable

from .datarobot_client import (
    UserBlueprint,
    ParamValuePair,
    UserBlueprintTaskArgumentDefinition,
    UserBlueprintTaskData,
    UserBlueprintTaskDefinition,
    ColnameAndType,
    UserBlueprintsValidateTaskParameter,
    UserBlueprintHexColumnNameLookupEntry,
)

from .dynamic import DynamicClass
from .exceptions import TaskFrozenError, CycleException, CustomTaskNotFound
from .friendly_repr import FriendlyRepr
from .user_blueprint_helper import UserBlueprintHelper
from .utils import (
    Colors,
    HiddenList,
    TaskOutputMethod,
    colorize,
    documentation,
    refresh_cached,
    represents_input_data,
)
from .workshop_interface import WorkshopInterface

TaskDict = Dict[str, Union[dict, str, List[str], List[Dict]]]
TaskParametersDict = Dict[str, Optional[Any]]

_COLNAME_PARAM = "cn"
_COLNAMES_PARAM = "cns"
_SELECT_COLNAME_METHOD = "method"


def hex_or_identity(task: Union[str, Task]) -> str:
    """
    Convenience method for retrieving the identifier for a task.
    This is a hex value if it's a Task and itself if an input data type.

    Parameters
    ----------
    task: str or Task
        Either a string of type BlueprintColumnClass or Task object

    Returns
    -------
    str
        id which represents a task- either a hex or an input data type
    """
    return str(task.task_id if isinstance(task, Task) else task)


class Task(FriendlyRepr, DynamicClass):
    """
    Provides a mechanism to specify how to transform or provide predictions for input
    data.

    Should not be used directly. Use one of: `Workshop.Task`, `Workshop.Tasks.*`,
    `Workshop.CustomTask`, or `Workshop.CustomTasks.*` instead.

    Parameters
    ----------
    workshop: WorkshopInterface
        Reference to the workshop
    task_code: string
        Task Code
    output_method: string
        starting with member of TaskMethodShort and an optional set of arguments
    task_parameters: TaskParameters
        A class with allowed params as attributes.
    output_method_parameters: List[ParamValuePair]
        A list of parameters and values to the operation
        e.g. `[ParamValuePair(param='add_orig_colname', value=True)]`
    x_transformations: List[ParamValuePair]
        A list of ParamValuePair representing transformations to be applied
        to the input data, X.
    y_transformations: List[ParamValuePair]
        A list of ParamValuePair representing transformations to be applied
        to the input targets, Y.
    freeze: boolean
        Whether or not to freeze the vertex - i.e. hexes, model strings, and inputs are cached.
    original_id: str
        An id to associate with this task which will be used when serializing the task.
    custom_task_id: str
        An id representing the custom task content- only applicable if the task is "Custom"
    version: str or None
        An id representing the custom task version content- only applicable if the task is "Custom"
    """

    _attr_to_refresh = ["_cached_input_names", "_cached_hex"]

    def __init__(
        self,
        workshop: WorkshopInterface,
        task_code: str,
        output_method: Optional[str] = None,
        task_parameters: Union[List[ParamValuePair], dict] = None,
        output_method_parameters: Union[List[ParamValuePair], dict] = None,
        x_transformations: Union[List[ParamValuePair], dict] = None,
        y_transformations: Union[List[ParamValuePair], dict] = None,
        freeze: bool = False,
        original_id: Optional[str] = None,
        custom_task_id: Optional[str] = None,
        version: Optional[str] = None,
        hex_column_name_lookup: Optional[
            List[UserBlueprintHexColumnNameLookupEntry]
        ] = None,
    ):
        """
        Parameters
        ----------
        workshop: WorkshopInterface
            Reference to the workshop
        task_code: string
            Task Code
        output_method: string
            starting with member of TaskMethodShort and an optional set of arguments
        task_parameters: List[ParamValuePair] or dict
            A list of ParamValue, e.g. `[ParamValuePair(param='n', value=100)]
            Or a dictionary where keys are params and values are values.
        output_method_parameters: List[ParamValuePair] or dict
            A list of parameters and values to the operation
            e.g. `[ParamValuePair(param='add_orig_colname', value=True)]
            Or a dictionary where keys are params and values are values.
        x_transformations: List[ParamValuePair] or dict
            A list of ParamValuePair representing transformations to be applied
            to the input data, X. Or a dictionary where keys are params and values are values.
        y_transformations: List[ParamValuePair] or dict
            A list of ParamValuePair representing transformations to be applied
            to the input targets, Y. Or a dictionary where keys are params and values are values.
        freeze: boolean
            Whether or not to freeze the vertex - i.e. hexes, model strings, and inputs are cached.
        original_id: str
            An id to associate with this task which will be used when serializing the task.
        custom_task_id: str
            An id representing the custom task content- only applicable if the task is "Custom"
        version: str or None
            An id representing the custom task version content- only applicable if the task is
            "Custom"
        hex_column_name_lookup: List[UserBlueprintHexColumnNameLookupEntry], (Default=None)
            Extra columns which exist in the User Blueprint, to be decoded
        """
        self._inputs = []
        self._workshop = workshop
        self._task_definitions = workshop._task_definitions_by_task_code
        self._custom_task_definitions = workshop._custom_tasks_by_id
        self._custom_task_version_definitions = workshop._custom_tasks_by_version_id
        self._input_name_lookup = workshop._input_name_lookup
        self._task_code = task_code
        self._output_method = output_method
        self._selected_features: List[ColnameAndType] = []
        self._exclude_features: bool = False
        self._dynamic_generation_exceptions = []
        self._hex_column_name_lookup = (
            {
                item.hex: ColnameAndType(
                    colname=item.colname, hex=item.hex, type="UNKNOWN"
                )
                for item in hex_column_name_lookup
            }
            if hex_column_name_lookup is not None
            else {}
        )
        self.version = None
        self._versions = []

        # Auto-convert dictionary formatted arguments
        if isinstance(task_parameters, dict):
            task_parameters = self._to_param_values_from_dict(task_parameters)
        if isinstance(output_method_parameters, dict):
            output_method_parameters = self._to_param_values_from_dict(
                output_method_parameters
            )
        if isinstance(x_transformations, dict):
            x_transformations = self._to_param_values_from_dict(x_transformations)
        if isinstance(y_transformations, dict):
            y_transformations = self._to_param_values_from_dict(y_transformations)

        version_not_found = False
        version_id = None

        if task_parameters:
            task_parameter_value_lookup = {tp.param: tp.value for tp in task_parameters}
            if custom_task_id is None and version is None:
                if "version_id" in task_parameter_value_lookup:
                    version_id = task_parameter_value_lookup["version_id"]
                    if version_id in self._custom_task_version_definitions:
                        version = version_id
                    else:
                        version_not_found = True

            # TODO: Remove reference to internal parameter
            hex_colname = task_parameter_value_lookup.get(_COLNAME_PARAM)
            if hex_colname is not None:
                self._selected_features = self._decode_colnames([hex_colname])

            hex_colnames = task_parameter_value_lookup.get(_COLNAMES_PARAM)
            if hex_colnames is not None:
                self._selected_features = self._decode_colnames(hex_colnames)

            self._exclude_features = (
                task_parameter_value_lookup.get(_SELECT_COLNAME_METHOD, "include")
                == "exclude"
            )

        if custom_task_id is None and version is not None:
            custom_task_id = self._custom_task_version_definitions[
                version
            ].custom_task_id

        if custom_task_id is not None:
            if custom_task_id not in self._custom_task_definitions:
                raise CustomTaskNotFound(
                    "Supplied custom task id was not found. Call `<Workshop>.refresh()` "
                    "if the custom task was added since creation of the `Workshop` instance."
                )

        self._task_definition: UserBlueprintTaskDefinition = (
            self._task_definitions.get(task_code, None)
            if custom_task_id is None
            else self._custom_task_definitions[custom_task_id]
        )

        # Ensure we handle backwards compatibility / execution of unsupported tasks.
        if self._task_definition is None:
            self._task_definition = (
                (UnavailableCustomTaskDefinition(task_code, version_id))
                if version_not_found and version_id is not None
                else (UnsupportedTaskDefinition(task_code))
            )

        self._task_parameter_definitions = self._task_definition.arguments
        self._task_parameter_definitions_by_key: Dict[
            str, UserBlueprintTaskArgumentDefinition
        ] = {entry.key: entry.argument for entry in self._task_definition.arguments}
        self._task_parameter_definitions_name_to_key: Dict[str, str] = {
            entry.argument.name: entry.key for entry in self._task_parameter_definitions
        }

        # Custom Task Related Initialization
        self.custom_task_id = custom_task_id
        if self.custom_task_id is not None:
            self.version = (
                version
                if version is not None
                else self._task_definition.custom_task_versions[0].id
            )

        if self._output_method is None:
            self._output_method = (
                TaskOutputMethod.PREDICT
                if (TaskOutputMethod.PREDICT in self._task_definition.output_methods)
                else TaskOutputMethod.TRANSFORM
            )

        # Dynamically build classes and methods based on the task parameters
        self._dynamic_initialization(task_parameters)

        self._output_method_parameters: List[ParamValuePair] = (
            output_method_parameters or []
        )

        # Direct data transformations
        # Ensure we properly represent x and y transformations as a dictionary
        self._x_transformations: List[ParamValuePair] = (
            copy.deepcopy(x_transformations) if x_transformations is not None else []
        )
        self._y_transformations: List[ParamValuePair] = (
            copy.deepcopy(y_transformations) if y_transformations is not None else []
        )

        # In case we'd like to map back to an original id for identification
        self.original_id = original_id

        # For performance, there are aspects we can cache
        self._frozen = freeze
        self._cached_hex = None
        self._cached_input_names = None

    def validate_task_parameters(
        self, color=True, only_output_errors=False, to_stdout=True
    ) -> List[UserBlueprintsValidateTaskParameter]:
        """Perform task validation to ensure all parameters and their values are valid."""
        task_parameter_dict = self._task_parameters_dict

        with self._workshop.auto_rate_limiter:
            result = UserBlueprint.validate_task_parameters(
                output_method=self.output_method,
                task_code=self.task_code,
                task_parameters=[
                    {"param_name": param, "new_value": value}
                    for param, value in task_parameter_dict.items()
                ],
            )

        # Dictionary using unique key
        error_lookup = {error.param_name: error for error in result.errors}

        for error in result.errors:
            error.param_name = self._task_parameter_definitions_by_key[
                error.param_name
            ].name

        lines = []
        if result.errors:
            lines = [
                "{} ({})".format(self.label, self.task_code_or_custom_task_code),
                "",
            ]

            if self.selected_features:
                lines += [self._formatted_selected_features()]
            else:
                for k in sorted(task_parameter_dict.keys()):
                    line = "{name} ({key}) = {value}".format(
                        name=self._task_parameter_definitions_by_key[k].name,
                        key=k,
                        value=task_parameter_dict[k],
                    )
                    if k in error_lookup:
                        additions = [
                            "  {}".format(error_lookup[k].message),
                            "    {}".format(line),
                            "      - Must be a '{}' parameter defined by: {}".format(
                                self._task_parameter_definitions_by_key[k].type,
                                self._task_parameter_definitions_by_key[k].values,
                            ),
                            "",
                        ]
                        if color:
                            for i, item in enumerate(additions):
                                additions[i] = colorize(item, Colors.RED)

                        lines += additions

        elif not only_output_errors:
            lines = [
                "{} ({})".format(self.label, self.task_code_or_custom_task_code),
                "",
            ]
            success = "All parameters valid!"
            if color:
                success = colorize(success, Colors.GREEN)
            lines += [success, ""]

        if lines and to_stdout:
            print("\n".join(lines))

        return HiddenList(result.errors)

    def documentation(self, auto_open=False) -> str:
        """
        Return a link to the documentation of the task.
        Optionally automatically open it in a browser window.
        """
        return documentation(self._task_definition, auto_open=auto_open)

    @property
    def task_code(self) -> str:
        """The unique code representing the class. E.g. "ABC"."""
        return self._task_code

    @property
    def task_code_or_custom_task_code(self) -> str:
        """Return the identifier of the task, either the task_code or the custom task id."""
        return (
            self._task_code
            if self.custom_task_id is None
            else ("{}_{}".format(self._task_code, self.custom_task_id))
        )

    @property
    def label(self) -> str:
        """Build the human-friendly label to describe the task."""
        if self.selected_features:
            return "{}, {}".format(
                self._task_definition.label, self._formatted_selected_features()
            )
        return self._task_definition.label

    @property
    def description(self) -> str:
        """The description of the task."""
        return self._task_definition.description

    @property
    def inputs(self) -> List[str]:
        """A list of the inputs to the task."""
        return self._inputs

    @property
    def output_method(self) -> str:
        """The output method of the task."""
        return self._output_method

    def get_task_parameter_by_name(self, name: str) -> Optional[Any]:
        """Retrieve the value of a task parameter by it's human-readable name."""
        key = self._task_parameter_definitions_name_to_key.get(name)
        return self._task_parameters_dict.get(
            key, self._task_parameter_definitions_by_key[key].default
        )

    @property
    def is_estimator(self) -> bool:
        """Whether or not the task can be used for prediction (stacked or otherwise)."""
        return TaskOutputMethod.is_estimator(self._output_method)

    @property
    def using_default_output_method(self) -> bool:
        """Return whether the task is using its default output method."""
        if self.is_estimator:
            return self.output_method == TaskOutputMethod.PREDICT
        return self.output_method == TaskOutputMethod.TRANSFORM

    @classmethod
    def deserialize(
        cls,
        workshop: WorkshopInterface,
        task_dict: TaskDict,
        original_id: Optional[str] = None,
    ) -> Task:
        """
        Take in a dictionary specifying how the task should be constructed,
        and construct an instance of the Task.
        """
        task = cls(
            workshop=workshop,
            task_code=task_dict["task_code"],
            output_method=task_dict["output_method"],
            task_parameters=cls._to_param_values_from_list_dict(
                task_dict["task_parameters"]
            ),
            output_method_parameters=cls._to_param_values_from_list_dict(
                task_dict["output_method_parameters"]
            ),
            x_transformations=cls._to_param_values_from_list_dict(
                task_dict["x_transformations"]
            ),
            y_transformations=cls._to_param_values_from_list_dict(
                task_dict["y_transformations"]
            ),
            original_id=original_id,
        )
        task._inputs = task_dict["inputs"]
        return task

    @classmethod
    def from_user_blueprint_task(
        cls,
        workshop: WorkshopInterface,
        task_data: UserBlueprintTaskData,
        original_id: str = None,
        hex_column_name_lookup: Optional[
            List[UserBlueprintHexColumnNameLookupEntry]
        ] = None,
    ) -> Task:
        """Instantiate a `Task` using `UserBlueprintTaskData` from a `UserBlueprint`."""
        task = cls(
            workshop=workshop,
            task_code=task_data.task_code,
            output_method=task_data.output_method,
            task_parameters=task_data.task_parameters,
            output_method_parameters=task_data.output_method_parameters,
            x_transformations=task_data.x_transformations,
            y_transformations=task_data.y_transformations,
            original_id=original_id,
            hex_column_name_lookup=hex_column_name_lookup,
        )
        task._inputs = task_data.inputs
        return task

    def serialize(
        self, map_ids: Optional[Dict[str, Union[str, Task]]] = None, minify: bool = True
    ) -> TaskDict:
        """
        Convert the instance of the Task into the serialized form of a dictionary.
        """
        if map_ids is None:
            input_values = [
                i.original_or_task_id if isinstance(i, Task) else i for i in self.inputs
            ]
        else:
            input_values = sorted(Task._build_input_mapping(self.inputs, map_ids))

        task_parameters_dict = (
            self._minified_task_parameters if minify else self._task_parameters_dict
        )

        return {
            "inputs": input_values,
            "task_code": self._task_code,
            "task_parameters": self._to_list_from_dict(task_parameters_dict),
            "output_method": self._output_method,
            "output_method_parameters": self._to_list_dict_from_list_param_value(
                self._output_method_parameters
            ),
            "x_transformations": self._to_list_dict_from_list_param_value(
                self._x_transformations
            ),
            "y_transformations": self._to_list_dict_from_list_param_value(
                self._y_transformations
            ),
        }

    @property
    def task_id(self) -> str:
        """
        Produces a hex which will be identical for two Task with
        the same inputs, task string, and task type.
        """
        if self._is_frozen and self._cached_hex:
            return self._cached_hex
        self._ensure_no_cycles()
        stringified = json.dumps(self.serialize())
        to_be_hashed = str(stringified).encode("utf-8")
        self._cached_hex = str(hashlib.md5(to_be_hashed).hexdigest())
        return self._cached_hex

    @property
    def original_or_task_id(self) -> str:
        """
        If the task has an id associated with it manually, via `original_id`, return
        this id, otherwise return the generated `self.task_id`.
        """
        return self.original_id if self.original_id is not None else self.task_id

    @property
    def input_names(self) -> List[str]:
        """Ordered representation of the input ids (things like 'NUM', '1', or <hex>)"""
        if self._is_frozen and self._cached_input_names:
            return self._cached_input_names
        self._cached_input_names = sorted([hex_or_identity(i) for i in self._inputs])
        return self._cached_input_names

    @property
    def selected_features(self) -> List[ColnameAndType]:
        """If specific features are selected, return them."""
        return self._selected_features

    @property
    def versions(self):
        """For Custom Tasks Only: The possible versions for a given task."""
        return self._versions

    # Methods below this point are not meant to be called directly by users.

    def __friendly_repr__(self) -> str:
        inputs = [
            "{} ({})".format(i.label, i.task_code_or_custom_task_code)
            if isinstance(i, Task)
            else "{} Data".format(self._input_name_lookup[i])
            for i in self._inputs
        ]
        sections = [
            "{} ({})".format(self.label, self.task_code_or_custom_task_code),
            "",
            "Input Summary: {}".format(" | ".join(inputs) if inputs else "(None)"),
            "Output Method: {}".format(
                TaskOutputMethod.get_variable_name(self.output_method)
            ),
        ]

        if self.selected_features:
            sections += [self._formatted_selected_features()]

        task_parameters_dict = self._task_parameters_dict
        if task_parameters_dict:
            sections += [
                "\n".join(["", "Task Parameters:"]),
                "  "
                + "\n  ".join(
                    [
                        "{} ({}) = {}".format(
                            self._task_parameter_definitions_by_key[key].name,
                            key,
                            task_parameters_dict[key].__repr__(),
                        )
                        if self.custom_task_id is None or key != "version_id"
                        else ("version_id (version_id) = {}".format(self.version))
                        for key in sorted(task_parameters_dict.keys())
                    ]
                ),
            ]

        return "\n".join(sections)

    def _dynamic_initialization(self, task_parameters: List[ParamValuePair]):
        """
        Dynamically build classes and functions to allow a first-class API-driven experience.
        """
        # Update based on task parameters, and fill defaults
        task_parameters: List[ParamValuePair] = task_parameters or []
        task_parameter_keys = set([tp.param for tp in task_parameters])

        task_parameter_value_lookup = {tp.param: tp.value for tp in task_parameters}

        class TaskParameters(object):
            def __init__(zelf):
                zelf._modified_keys = set()

            def __setattr__(zelf, k, v):
                # It is imperative this is `is not` instead of `!=` as `!=` will perform
                # type coercion and `0 == False` which, when `0` is a select index, this should
                # fail, not pass.
                if k != "_modified_keys" and getattr(zelf, k, v) is not v:
                    zelf._modified_keys.add(k)
                if k == "version_id" and self.version != v:
                    # If setting this directly, keep version updated...
                    self.version = v
                super(TaskParameters, zelf).__setattr__(k, v)

        self.task_parameters = TaskParameters()
        keys_to_keep = set()
        arg_key_to_argument_pairs = {}

        arguments = UserBlueprintHelper.get_arguments(self._task_definition.arguments)
        for entry in arguments:
            key = entry.key
            value = entry.argument
            arg_key_to_argument_pairs[key] = value
            if key in task_parameter_keys:
                self._set_task_parameter(key, task_parameter_value_lookup[key])
                keys_to_keep.add(key)
            else:
                self._set_task_parameter(key, value.default)

        TaskParameters.__doc__ = UserBlueprintHelper.get_docstring_for_arguments(
            arguments, title=self._task_definition.label
        )

        def task_parameters_repr(zelf):
            sorted_keys = sorted(
                zelf.__dict__.keys(),
                key=lambda k: (
                    self._task_parameter_definitions_by_key[k].name
                    if k in self._task_parameter_definitions_by_key
                    else k
                ),
            )
            return "\n".join(
                [
                    "{} ({}) = {}".format(
                        self._task_parameter_definitions_by_key[k].name,
                        k,
                        zelf.__dict__[k].__repr__(),
                    )
                    for k in sorted_keys
                    if k != "_modified_keys"
                ]
            )

        TaskParameters.__repr__ = task_parameters_repr
        self._dynamic_set_parameter_methods(arg_key_to_argument_pairs)

        self.task_parameters._modified_keys = set(keys_to_keep)

        if self.version:
            self._versions = build_custom_task_versions(
                task_definition=self._task_definition
            )

    def _dynamic_set_parameter_methods(
        self, arg_key_to_argument_pairs: Dict[str, UserBlueprintTaskArgumentDefinition]
    ):
        """
        Dynamically generate `set_task_parameters` and `set_task_parameters_by_name.
        """
        # Rename params that are also keywords (e.g. 'is', 'in', 'and', 'or', etc.)
        lookup = {
            k + "_" if iskeyword(k) else k: v
            for k, v in arg_key_to_argument_pairs.items()
        }

        params_by_name = {
            lookup[key].name: lookup[key]
            for key in lookup.keys()
            if re.match("^[_a-zA-Z0-9]+$", lookup[key].name)
        }

        def get_parameter_docstring(
            argument: UserBlueprintTaskArgumentDefinition, key: Optional[str] = None
        ):
            field_name = argument.name if key is None else key
            template = "{field_name}: {type}, (Default={default})\n"
            template += "    Possible Values: {values}\n"
            return template.format(
                field_name=field_name,
                type=argument.type,
                default=argument.default.__repr__(),
                values=argument.values,
            )

        # Add `set_task_parameters`
        self._dynamic_generation_exceptions += (
            self._wrap_function_with_kwargs_and_add_to_class(
                fn_name="set_task_parameters",
                fn_to_wrap=self._set_task_parameters,
                kwargs_dict=lookup,
                docstring=UserBlueprintHelper.get_docstring_for_lines(
                    lines=[
                        get_parameter_docstring(lookup[key], key=key)
                        for key in sorted(lookup.keys())
                    ],
                    title="Set parameters by their key",
                    minified=False,
                ),
            )
        )

        # Add `set_task_parameters_by_name`
        self._dynamic_generation_exceptions += (
            self._wrap_function_with_kwargs_and_add_to_class(
                "set_task_parameters_by_name",
                self._set_task_parameters_by_name,
                kwargs_dict=params_by_name,
                docstring=UserBlueprintHelper.get_docstring_for_lines(
                    lines=[
                        get_parameter_docstring(params_by_name[name])
                        for name in sorted(params_by_name.keys())
                    ],
                    title="Set parameters by their name",
                    minified=False,
                ),
            )
        )

    def _map_inputs(self, mapping: Dict[str, Task]) -> Task:
        """
        Set the inputs according to a provided map from one set of ids or values to another.

        Useful for changing the order of the nodes (e.g. topological sorting).
        """
        self._inputs = Task._build_input_mapping(self._inputs, mapping)
        return self

    def _formatted_selected_features(self) -> str:
        """A human readable representation of the selected features being used."""
        prefix = "{verb} {noun}: ".format(
            verb="Select" if not self._exclude_features else "Exclude",
            noun="Feature" if len(self.selected_features) == 1 else "Features",
        )
        return prefix + ", ".join(
            ["'{}'".format(f.colname) for f in self.selected_features]
        )

    def _domain(self) -> str:
        """The hostname to use to access documentation."""
        return UserBlueprint._client.domain

    def _decode_colnames(self, hex_colnames: List[str]):
        """Try to decode the column names if possible."""
        if hex_colnames is None:
            return []
        return [
            self._colname_and_type_from_hex(
                self._workshop, hex_colname, self._hex_column_name_lookup
            )
            for hex_colname in hex_colnames
        ]

    @property
    def _task_parameters_dict(self) -> TaskParametersDict:
        """Serialize the TaskParameters instance into a dictionary."""
        task_parameters_dict = copy.deepcopy(self.task_parameters.__dict__)
        drop_keys = (
            set(task_parameters_dict.keys()) - self.task_parameters._modified_keys
        )
        for key in drop_keys:
            task_parameters_dict.pop(key, None)
        if self.custom_task_id is not None:
            task_parameters_dict["version_id"] = (
                self.version
                if self.version is not None
                else "latest_{}".format(self.custom_task_id)
            )
        return task_parameters_dict

    @property
    def _task_parameter_keys(self) -> List[str]:
        """Get a list of the task parameters."""
        return sorted(self._task_parameters_dict.keys())

    @property
    def _minified_task_parameters(self) -> TaskParametersDict:
        """Get all the TaskParameters which are not using their default values."""
        parameters = {}
        for k, v in self._task_parameters_dict.items():
            if self.custom_task_id is not None and k == "version_id":
                parameters[k] = v
                continue

            if k not in self._task_parameter_definitions_by_key:
                continue
            definition = self._task_parameter_definitions_by_key[k]
            default_value = definition.default

            def equivalent_to_default(
                value, default=default_value, allow_coercion=True
            ):
                if allow_coercion and value == default:
                    return True
                return any([value is default, str(value) == str(default)])

            def value_is_index_and_value(value):
                """
                Try to see if `value` is an index...
                But ensure we don't coerce a value existing in the possible values to be an index.
                """
                values = definition.values
                if isinstance(value, (str, int)) and str(value) not in [
                    str(val) for val in values
                ]:
                    try:
                        index = int(value)
                        if index < len(values):
                            return True, values[index]
                    except (ValueError, TypeError):
                        pass
                return False, None

            # Handle indexing into select
            if definition.type in ["select", "selectgrid"]:
                # Try to see if `v` is an index...
                is_index, val = value_is_index_and_value(v)
                if is_index and equivalent_to_default(val, allow_coercion=False):
                    continue

                # Try to see if `default` is an index...
                is_index, val = value_is_index_and_value(default_value)
                if is_index and equivalent_to_default(
                    v, default=val, allow_coercion=False
                ):
                    continue

            if equivalent_to_default(
                v, allow_coercion=definition.type not in ["select", "selectgrid"]
            ):
                continue

            parameters[k] = v

        return parameters

    def _set_task_parameters_by_name(self, **parameters) -> Task:
        """Set task parameters by human-readable names in bulk."""
        for name, value in parameters.items():
            self._set_task_parameter_by_name(name, value)
        return self

    def _set_task_parameters(self, **parameters) -> Task:
        """Set task parameters in bulk."""
        for k, v in parameters.items():
            if k.endswith("_") and iskeyword(k):
                k = k[:-1]
            self._set_task_parameter(k, v)
        return self

    def _set_task_parameter_by_name(self, name, value) -> Task:
        """Set a specific task parameter by a human-readble name."""
        key = self._task_parameter_definitions_name_to_key.get(name)
        self._set_task_parameter(key, value)
        return self

    def _set_task_parameter(self, key, value) -> Task:
        """Set a specific task parameter."""
        if self._is_frozen:
            raise TaskFrozenError(
                "This task is frozen and cannot be modified. Call `<Task>._unfreeze() first."
            )
        setattr(self.task_parameters, key, value)
        return self

    @staticmethod
    def _to_list_from_dict(input_dict: dict) -> List[dict]:
        """
        Turn a dynamically-keys dictionary representing parameters and values into
        the list of static keyed "param" and "value" dictionaries.
        """
        return sorted(
            [{"param": p, "value": v} for p, v in input_dict.items()],
            key=lambda x: x["param"],
        )

    @staticmethod
    def _to_list_dict_from_list_param_value(
        list_param_value: List[ParamValuePair],
    ) -> List[dict]:
        """
        Convert a list of `ParamValuePair` into a list of dictionaries with "param" and "value"
        keys with their respective values.
        """
        return sorted(
            [param_value.as_json() for param_value in list_param_value],
            key=lambda x: x["param"],
        )

    @staticmethod
    def _to_param_values_from_list_dict(
        input_list: List[Dict[str, str]]
    ) -> List[ParamValuePair]:
        """
        Convert a list of "param", "value" keyed dictionary
        into a list of `ParamValuePair`
        """
        return [ParamValuePair(**p_v) for p_v in input_list]

    @staticmethod
    def _to_param_values_from_dict(input_dict: Dict[str, str]) -> List[ParamValuePair]:
        """
        Convert a dict of parameters and values into a list of `ParamValuePair`
        """
        return [ParamValuePair(p, v) for p, v in input_dict.items()]

    @staticmethod
    def _build_input_mapping(
        inputs, mapping: Optional[Dict[str, Union[str, Task]]] = None
    ) -> List[str]:
        """
        Builds a map from one set of inputs to another, based on a provided map,
        while preserving any inputs which do not have a mapping. (e.g. input types).

        Useful for transforming index-based inputs to actual vertices.

        For example:
        ['NUM', '1', '2'] -> ['NUM', <Task:0x1>, <Task:0x2>]

        Parameters
        ----------
        inputs: dict(string, string or Task)
            Inputs to map to new keys or values.
        mapping: dict or None (Default=None)
            By default, there is no mapping, so there's nothing to change. But, if a
            dictionary is passed in, map the values accordingly.

        Returns
        -------
        dict or list
        """
        if mapping is None:
            return inputs
        return [mapping.get(hex_or_identity(vertex), vertex) for vertex in inputs]

    def _ensure_no_cycles(self, seen=None):
        """Descend into the inputs of the current task to ensure there are no cycles."""
        if seen is None:
            seen = {id(self)}
        task_inputs = [inp for inp in self._inputs if isinstance(inp, Task)]
        for inp in task_inputs:
            if id(inp) in seen:
                # There's a cycle, so we can't properly resolve the hex
                raise CycleException("Cycle found: tasks cannot contain cycles.")
            else:
                seen_copy = copy.deepcopy(seen)
                seen_copy.add(id(inp))
                inp._ensure_no_cycles(seen_copy)
                seen_copy.remove(id(inp))

    def _refresh_cached(self, attr_to_exclude=None):
        """Refresh cached attributes (set to None) of Task.

        Parameters
        ----------
        attr_to_exclude:
            A list of attributes which are not to be refreshed.
        """
        refresh_cached(self, self._attr_to_refresh, attr_to_exclude)

    def _freeze(self):
        """Internal only: freeze the task and its inputs for performance."""
        if self._is_frozen:
            return
        self._refresh_cached()
        for i in self._inputs:
            if isinstance(i, Task):
                i._freeze()
        self._frozen = True

    def _unfreeze(self):
        """Internal only: unfreeze the task and its inputs for modification."""
        for i in self._inputs:
            if isinstance(i, Task):
                i._unfreeze()
        self._frozen = False

    def _assert_unfrozen(self):
        """Ensure the task is frozen, or raise an exception."""
        if self._is_frozen:
            raise TaskFrozenError(
                "This task is frozen and cannot be modified. Call `<Task>._unfreeze() first."
            )

    def __setattr__(self, key, value):
        """Ensure the task is not frozen before setting."""
        if key != "_frozen" and key not in self._attr_to_refresh:
            self._assert_unfrozen()
        super(Task, self).__setattr__(key, value)

    def __hash__(self):
        """The hash representation of the task."""
        return int(self.task_id, 16)

    def __eq__(self, other):
        """A way to determine if two tasks are identical in parameters and input."""
        if not isinstance(other, Task):
            return False
        return self.serialize() == other.serialize()

    @property
    def _is_frozen(self):
        return getattr(self, "_frozen", False)

    @classmethod
    def _colname_and_type_from_name(
        cls, workshop: WorkshopInterface, feature_name: str
    ) -> ColnameAndType:
        return workshop._colname_and_type_colname_lookup.get(feature_name)

    @classmethod
    def _colname_and_type_from_hex(
        cls,
        workshop: WorkshopInterface,
        feature_hex: str,
        alternate_lookup: Optional[Dict[str, ColnameAndType]] = None,
    ) -> ColnameAndType:
        subject = str(feature_hex)
        feature = workshop._colname_and_type_hex_lookup.get(subject)
        if feature is not None:
            return feature
        fallback = ColnameAndType(hex=subject, colname="_UNKNOWN_", type="UNKNOWN")
        if alternate_lookup is not None:
            fallback = alternate_lookup.get(subject, fallback)
        return fallback

    def __call__(self, *inputs, replace_inputs=False):
        """
        Calling a Task saves a reference to the provided input arguments
        as inputs to the caller, creating a directed graph.

        Any Task can be passed as input to a BlueprintGraph to create a full
        blueprint. At which point, it can be ensured that it's a valid DAG and be topologically
        sorted, along with other necessary verification of criteria necessary to be a valid
        DataRobot blueprint.

        The inputs may be one of: an input data type string, another Task,
        or a Blueprint.

        Parameters
        ----------
        functional_members: list(Task or string)
            Note: Any string passed must be a member of BlueprintColumnClass
        replace_inputs: bool
            Whether to replace the current inputs, instead of append to them
        """
        new_inputs = []
        for m in inputs:
            if isinstance(m, Task) or isinstance(m, str) and represents_input_data(m):
                new_inputs.append(m)
            else:
                raise TypeError(
                    "Inputs must either be a Task, or a string of the TaskInput"
                )
        if replace_inputs:
            self._inputs = new_inputs
        else:
            self._inputs += new_inputs
        return self


class FeatureSelection(Task):
    """
    A special type of task that selects features from project data.

    Please note: This feature is unreleased and may not be available on your account.
    """

    def __init__(
        self,
        workshop: WorkshopInterface,
        feature_names: Iterable[Union[str, Feature]],
        task_code="MCPICK",
        exclude: bool = False,
        **kwargs,
    ):
        # Convert any `Feature`s to just the colname
        feature_names = [
            f.selected_features[0].colname if isinstance(f, Task) else f
            for f in feature_names
        ]

        # TODO: Remove the need to reference a specific task
        super(FeatureSelection, self).__init__(
            workshop=workshop,
            task_code=task_code,
            task_parameters=self.build_task_parameters(
                workshop, feature_names, exclude=exclude
            ),
            **kwargs,
        )
        # Ensure we add the proper input
        data_types = sorted(
            set(
                self._colname_and_type_from_name(workshop, feature_name).type
                for feature_name in feature_names
            )
        )
        if len(data_types) > 1:
            raise ValueError(
                "Multiple data types found: {}, only 1 allowed.".format(data_types)
            )
        self(data_types[0])

    def select_features(self, feature_names: Iterable[str]) -> Task:
        """Choose a specific feature, and only take input from this feature."""
        task_parameters = self.build_task_parameters(self._workshop, feature_names)
        return self._set_task_parameters(
            **{tp.param: tp.value for tp in task_parameters}
        )

    @classmethod
    def build_task_parameters(
        cls,
        workshop: WorkshopInterface,
        feature_names: Iterable[str],
        exclude: bool = False,
    ) -> List[ParamValuePair]:
        """Build the task parameters for a FeatureSelection."""
        for feature_name in feature_names:
            if cls._colname_and_type_from_name(workshop, feature_name) is None:
                raise ValueError(
                    'The feature "{}" does not exist in this project.'.format(
                        feature_name
                    )
                )

        return [
            ParamValuePair(
                param=_COLNAMES_PARAM,
                value=[
                    str(cls._colname_and_type_from_name(workshop, feature_name).hex)
                    for feature_name in feature_names
                    if cls._colname_and_type_from_name(workshop, feature_name)
                    is not None
                ],
            ),
            ParamValuePair(
                param=_SELECT_COLNAME_METHOD, value="exclude" if exclude else "include"
            ),
        ]


class Feature(FeatureSelection):
    """A special type of task that selects a feature from project data."""

    def __init__(
        self,
        workshop: WorkshopInterface,
        feature_name: str,
        task_code="SCPICK",
        **kwargs,
    ):
        super(Feature, self).__init__(
            workshop=workshop,
            feature_names=[feature_name],
            task_code=task_code,
            **kwargs,
        )

    @classmethod
    def build_task_parameters(
        cls,
        workshop: WorkshopInterface,
        feature_names: List[str],
        exclude: bool = False,
    ) -> List[ParamValuePair]:
        """Build the task parameters for a Feature."""
        if len(feature_names) > 1:
            raise ValueError("Only a single feature is allowed.")

        feature_name = feature_names[0]
        feature = cls._colname_and_type_from_name(workshop, feature_name)
        if feature is None:
            raise ValueError(
                'The feature "{}" does not exist in this project.'.format(feature_name)
            )

        hex_feature = str(feature.hex)
        return [ParamValuePair(param=_COLNAME_PARAM, value=hex_feature)]


class UnsupportedTaskDefinition(UserBlueprintTaskDefinition):
    """This task is not supported, but can still be used."""

    def __init__(self, task_code: str):
        print('Note: Unsupported task initialized: "{}"'.format(task_code))
        super(UnsupportedTaskDefinition, self).__init__(
            arguments=[],
            categories=[],
            description="This task is not supported, but can still be used.",
            icon=0,
            label=task_code,
            output_methods=[TaskOutputMethod.all()],
            task_code=task_code,
            time_series_only=False,
            url="",
            valid_inputs=[],
        )


class UnavailableCustomTaskDefinition(UserBlueprintTaskDefinition):
    """This custom task is not available."""

    def __init__(self, task_code: str, version_id: str):
        print(
            (
                'Note: Unavailable custom task: "{}", with version: "{}". '
                "This custom task may no longer exist, or may be inaccessible to you. "
                "Please request the owner of the blueprint to share this task with you."
            ).format(task_code, version_id)
        )
        super(UnavailableCustomTaskDefinition, self).__init__(
            arguments=[],
            categories=[],
            description="This task is not available to you.",
            icon=0,
            label="Unknown Custom Task",
            output_methods=[TaskOutputMethod.all()],
            task_code=task_code,
            time_series_only=False,
            url="",
            valid_inputs=[],
        )


def build_custom_task_versions(task_definition: UserBlueprintTaskDefinition):
    """
    Create a readable list of versions with id and label for each, for a custom task.

    Parameters
    ----------
    task_definition: UserBlueprintTaskDefinition

    Returns
    -------
    Versions
        The list of versions of a custom task.
    """

    class Versions(FriendlyRepr):
        __doc__ = UserBlueprintHelper.get_docstring_for_versions(
            task_definition.custom_task_versions,
            title=task_definition.label,
        )

        def __init__(self):
            for version in task_definition.custom_task_versions:
                setattr(self, version.label.replace(".", "_"), version.id)

        def __friendly_repr__(self):
            return UserBlueprintHelper.get_docstring_for_versions(
                task_definition.custom_task_versions,
                title=task_definition.label,
                minified=True,
            )

    return Versions()



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""
from typing import Any, Dict, List, Callable


class DynamicClass(object):
    """Allow a class to have new methods added dynamically."""

    def _wrap_function_with_kwargs_and_add_to_class(
        self,
        fn_name: str,
        fn_to_wrap: Callable,
        kwargs_dict: Dict[str, Any],
        docstring: str,
    ) -> List[Exception]:
        try:

            def dynamic_method(**kwargs):
                for k in kwargs.keys():
                    if k not in kwargs_dict:
                        raise TypeError(
                            "{}() got an unexpected keyword argument '{}'".format(
                                fn_name, k
                            )
                        )
                return fn_to_wrap(**kwargs)

            dynamic_method.__doc__ = docstring
            self._add_function_to_class(fn_name, dynamic_method)
            return []
        except Exception as e:
            print("Failed to dynamically generate `{}`".format(fn_name), e)
            return [e]

    def _add_function_to_class(self, fn_name: str, fn: Callable):
        """Add a specified bit of code as a function to the current class"""
        fn.__name__ = fn_name
        return setattr(self, fn_name, fn)



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""
from __future__ import annotations

from typing import Dict, Optional, List, Collection

from datarobot import UserBlueprint

from .datarobot_client import (
    UserBlueprintAvailableTasks,
    UserBlueprintTaskArgument,
    UserBlueprintTaskDefinition,
    ColnameAndType,
)

from datarobot_bp_workshop.auto_rate_limiter import AutoRateLimiter
from datarobot_bp_workshop.sharing import SharingInterface
from .user_blueprint_helper import UserBlueprintHelper


class WorkshopInterface(SharingInterface):
    """Interface to use to allow working with a Workshop in BlueprintGraph and Task."""

    def __init__(self, project_id=None, enforce_rate_limit=True):
        self._task_definitions_by_task_code: Dict[str, UserBlueprintTaskDefinition] = {}
        self._custom_task_definitions_by_task_code: Dict[
            str, List[UserBlueprintTaskDefinition]
        ] = {}
        self._custom_tasks_by_id: Dict[str, UserBlueprintTaskDefinition] = {}
        self._custom_tasks_by_version_id: Dict[str, UserBlueprintTaskDefinition] = {}
        self._task_argument_lookup: Dict[str, Dict[str:UserBlueprintTaskArgument]] = {}
        self._input_name_lookup: Dict[str, str] = {}
        self._available_tasks: Optional[UserBlueprintAvailableTasks] = None
        self._colname_and_type_colname_lookup: Dict[str, ColnameAndType] = {}
        self._colname_and_type_hex_lookup: Dict[str, ColnameAndType] = {}
        self._project_id: Optional[str] = project_id
        self._associated_user_blueprint_id: Optional[str] = None

        self.auto_rate_limiter: AutoRateLimiter = (
            AutoRateLimiter() if enforce_rate_limit else AutoRateLimiter(0)
        )

    @property
    def project_id(self):
        return self._project_id

    def initialize(
        self, user_blueprint_id: Optional[str] = None, clean: bool = False
    ) -> WorkshopInterface:
        """Must be implemented on inheriting class."""
        raise NotImplementedError()

    def _custom_task_key_string(self, custom_task_id: str) -> str:
        task_code = self._custom_tasks_by_id[custom_task_id].task_code
        return "{}_{}".format(task_code, custom_task_id)

    def has_all_version_ids(self, version_ids: Collection[str]) -> bool:
        """Return whether all version ids are already available."""
        if not len(version_ids):
            return True
        keys = set(self._custom_tasks_by_version_id.keys())
        if len(set(version_ids) - keys):
            return False
        for version_id in version_ids:
            task_code = self._custom_task_key_string(
                self._custom_tasks_by_version_id[version_id].custom_task_id
            )
            custom_tasks = getattr(self, "CustomTasks", None)
            if not hasattr(custom_tasks, task_code):
                return False
        return True

    def has_all_task_codes(self, task_codes: Collection[str]) -> bool:
        """Return whether all task_codes are already available."""
        if not len(task_codes):
            return True
        keys = set(self._task_definitions_by_task_code.keys())
        if len(set(task_codes) - keys):
            return False
        return True

    def ensure_definitions_for_user_blueprint(self, user_blueprint: UserBlueprint):
        """
        Retrieve definitions for built-in or custom tasks, if they are not available.
        """
        # If there are any custom tasks that we have not retrieved definitions for,
        # do so now.
        version_ids = UserBlueprintHelper.get_custom_task_version_ids(user_blueprint)
        if not self.has_all_version_ids(version_ids):
            self.initialize(user_blueprint_id=user_blueprint.user_blueprint_id)
        else:
            # If there are any built-in tasks that we have not retrieved definitions for,
            # do so now
            task_codes = UserBlueprintHelper.get_task_codes(user_blueprint)
            if not self.has_all_task_codes(task_codes):
                self.initialize(user_blueprint_id=user_blueprint.user_blueprint_id)



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

from __future__ import annotations

from typing import Any, Callable, Generic, List, TypeVar

from .datarobot_client import UserBlueprintTaskDefinition

from datarobot_bp_workshop.friendly_repr import FriendlyRepr
from datarobot_bp_workshop.task import Task, build_custom_task_versions
from datarobot_bp_workshop.user_blueprint_helper import UserBlueprintHelper
from datarobot_bp_workshop.utils import documentation, TaskOutputMethod
from datarobot_bp_workshop.workshop_interface import WorkshopInterface


class TaskFactory(object):
    def __init__(
        self,
        workshop: WorkshopInterface,
    ):
        self._workshop = workshop
        self.task_definition_lookup = workshop._task_definitions_by_task_code
        self.custom_task_definition_lookup = workshop._custom_tasks_by_id
        self.custom_task_version_definition_lookup = (
            workshop._custom_tasks_by_version_id
        )

    def get_key(self, task_definition: UserBlueprintTaskDefinition):
        if task_definition.custom_task_id is None:
            return task_definition.task_code

        return "{}_{}".format(task_definition.task_code, task_definition.custom_task_id)

    def __call__(self, task_definition: UserBlueprintTaskDefinition):
        """
        Parameters
        ----------
        task_definition: UserBlueprintTaskDefinition
        """
        arguments = UserBlueprintHelper.get_arguments(task_definition.arguments)

        # Important that this is an object, as we use __dict__
        class TaskParameters(FriendlyRepr):
            def __init__(zelf):
                for argument in arguments:
                    setattr(zelf, argument.key, argument.argument)

            def __friendly_repr__(self):
                return UserBlueprintHelper.get_docstring_for_arguments(
                    arguments, title=task_definition.label, minified=True
                )

        TaskParameters.__doc__ = UserBlueprintHelper.get_docstring_for_arguments(
            arguments, title=task_definition.label
        )

        class Helper(object):
            __doc__ = "\n".join(
                [
                    task_definition.label,
                    "",
                    task_definition.description,
                    "",
                    "Parameters",
                    "----------",
                    "output_method: string, one of ({}).".format(
                        ", ".join(
                            [
                                TaskOutputMethod.get_variable_name(m)
                                for m in task_definition.output_methods
                            ]
                        )
                    ),
                    "task_parameters: dict, which may contain:",
                ]
                + [
                    "\n  "
                    + "\n\n  ".join(
                        [
                            (
                                "{name} ({key}): {type}, (Default={default})"
                                "\n    Possible Values: {values}"
                            ).format(
                                name=a.argument.name,
                                key=a.key,
                                type=a.argument.type,
                                default=a.argument.default.__repr__(),
                                values=a.argument.values,
                            )
                            for a in UserBlueprintHelper.get_arguments(
                                task_definition.arguments
                            )
                        ]
                    )
                ]
            )

            description = task_definition.description
            label = task_definition.label
            task_code = task_definition.task_code

            def documentation(zelf, auto_open=False):
                return documentation(task_definition, auto_open)

            task_parameters = TaskParameters()

            def __call__(
                zelf,
                *inputs,
                output_method=None,
                task_parameters=None,
                output_method_parameters=None,
                x_transformations=None,
                y_transformations=None,
                freeze=False,
                version=None,
            ):
                return Task(
                    workshop=self._workshop,
                    task_code=task_definition.task_code,
                    output_method=output_method,
                    task_parameters=task_parameters,
                    output_method_parameters=output_method_parameters,
                    x_transformations=x_transformations,
                    y_transformations=y_transformations,
                    freeze=freeze,
                    custom_task_id=task_definition.custom_task_id,
                    version=version,
                )(*inputs)

            def __friendly_repr__(zelf):
                name = "{}: [{}{}]".format(
                    zelf.label,
                    task_definition.task_code,
                    ""
                    if task_definition.custom_task_id is None
                    else "_{}".format(task_definition.custom_task_id),
                )
                return "{} \n  - {}\n".format(
                    name, zelf.description or "(No description)"
                )

        tasks_class_attributes = dict(
            __doc__=Helper.__doc__,
            __call__=Helper.__call__,
            __friendly_repr__=Helper.__friendly_repr__,
            description=Helper.description,
            label=Helper.label,
            task_code=Helper.task_code,
            documentation=Helper.documentation,
            task_parameters=Helper.task_parameters,
        )

        if task_definition.custom_task_id:
            tasks_class_attributes["versions"] = build_custom_task_versions(
                task_definition
            )

        tasks_class_type = type(
            self.get_key(task_definition), (FriendlyRepr,), tasks_class_attributes
        )

        return tasks_class_type()


T = TypeVar("T")


class SimpleFactory(Generic[T]):
    """Build a data-driven class and instantiate it."""

    def __init__(self, field_items: List[T]) -> None:
        self.field_items = field_items

    def __call__(
        self,
        get_key: Callable[[T], str],
        get_value: Callable[[T], Any],
        get_type: Callable[[T], str],
        get_description: Callable[[T], str],
        title: str,
        class_name: str,
    ):
        field_docstring_lines = [
            "{}: {}"
            "\n  {}".format(
                get_key(field_item),
                get_type(field_item),
                get_description(field_item),
            )
            for field_item in self.field_items
        ]

        docstring = "\n".join(
            [title, "", "Parameters", "----------"] + field_docstring_lines
        )

        def __friendly_repr__(zelf):
            return (
                title
                + "\n\n"
                + "{}".format(
                    ", ".join(get_key(field_item) for field_item in self.field_items)
                )
            )

        class_attributes = dict(
            {
                get_key(field_item): get_value(field_item)
                for field_item in self.field_items
            },
            __doc__=docstring,
            __friendly_repr__=__friendly_repr__,
        )

        class_type = type(class_name, (FriendlyRepr,), class_attributes)
        return class_type()



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

import time


class AutoRateLimiter(object):
    """
    A safety feature to ensure users don't accidentally rate-limit themselves.

    This ensures that if less than (1 / max requests per second) time has passed
    since the last request, ensure at least this much time as passed before
    completing the request.

    This is specifically to protect users that are using this functionality
    asynchronously, as everything is serial by default.
    """

    ONE_BILLION = 1e9

    def __init__(self, max_requests_per_second: int = 45):
        """
        Note: DataRobot API allows max 50/s

        Parameters
        ----------
        max_requests_per_second: int
            Number of requests per second to allow

        """
        self.enforce_rate_limitation = max_requests_per_second > 0
        self.max_requests_per_second = max_requests_per_second
        self.safety_delay_ns = (
            1.0 / max_requests_per_second if self.enforce_rate_limitation else 0
        ) * self.ONE_BILLION

        self.last_request = time.time_ns()

    def __enter__(self):
        if not self.enforce_rate_limitation:
            return
        now = time.time_ns()
        since_last = now - self.last_request
        if since_last < self.safety_delay_ns:
            delay_ns = self.safety_delay_ns - since_last
            time.sleep(delay_ns * 1e-9)
            now += delay_ns
        self.last_request = now

    def __exit__(self, type, value, traceback):
        pass



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

from __future__ import annotations

from contextlib import contextmanager
from typing import Dict, List, Optional, Union

import networkx as nx
from datarobot import Project
from .datarobot_client import (
    UserBlueprint,
    UserBlueprintTask,
    UserBlueprintHexColumnNameLookupEntry,
)
from .datarobot_client import UserBlueprintTaskDefinition as TaskDefinition
from .datarobot_client import VertexContextItem
from networkx import DiGraph

from .exceptions import (
    UnsavedBlueprintException,
    CycleException,
    NoProjectSpecifiedException,
)
from .friendly_repr import FriendlyRepr
from .task import Task, hex_or_identity
from .utils import (
    Colors,
    TaskInput,
    TaskOutputMethod,
    colorize,
    refresh_cached,
    represents_input_data,
    Roles,
    PrettyString,
)
from .visualize import Visualize, DisplayGraphviz
from .workshop_interface import WorkshopInterface


class BlueprintGraph(FriendlyRepr):
    """
    Provides a mechanism to declare a set of tasks in the form of a directed acyclic graph.

    Should not be used directly. Use `Workshop.Blueprint` instead.

    Parameters
    ----------
    workshop: WorkshopInterface
        Reference to the workshop
    final_task_or_tasks: Task or list(task)
        Either:
            - the task to use as the final task in a blueprint, which will be inspected
              to find all inputs.
            - the list of tasks to be used in the BlueprintGraph

    name: str or None
        The name to give the Blueprint Graph, which will be used to label the UserBlueprint
        if saved to DataRobot.
    """

    def __init__(
        self,
        workshop: WorkshopInterface,
        final_task_or_tasks: Union[Task, List[Task]],
        name: Optional[str] = None,
    ):
        self._workshop = workshop
        self._frozen: bool = False
        self._cached_graph: Optional[DiGraph] = None
        self._task_definitions: Dict[
            str, TaskDefinition
        ] = workshop._task_definitions_by_task_code
        self._custom_task_definitions: Dict[
            str, TaskDefinition
        ] = workshop._custom_tasks_by_id
        self._custom_task_version_definitions: Dict[
            str, TaskDefinition
        ] = workshop._custom_tasks_by_version_id
        self._input_name_lookup: Dict[str, str] = workshop._input_name_lookup
        self.user_blueprint_id: Optional[str] = None
        if isinstance(final_task_or_tasks, Task):
            self.tasks = list(set(self._tasks_from_task(final_task_or_tasks).values()))
        else:
            self.tasks = final_task_or_tasks

        if any([not isinstance(v, Task) for v in self.tasks]):
            raise TypeError("`tasks` must be a list of Task objects")

        self.name = name
        self.vertex_context: List[VertexContextItem] = []
        self._user_blueprint: Optional[UserBlueprint] = None
        self.project_id = None

        # Extra information that may be available
        self.creator: Optional[str] = None
        self.last_modifier_name: Optional[str] = None

    def __call__(self, final_task: Task, name=None):
        """Assign a new final task for an existing BlueprintGraph."""
        self.tasks = list(set(self._tasks_from_task(final_task).values()))
        if name is not None:
            self.name = name

    def save(
        self,
        name=None,
        user_blueprint_id=None,
        save_as_new=False,
        validate_parameters=True,
        color=True,
    ) -> BlueprintGraph:
        """
        Save a BlueprintGraph as a User Blueprint in DataRobot.

        Returns
        -------
        BlueprintGraph
        """
        errors = []
        if validate_parameters:
            for task in self._traverse_dag():
                errors += task.validate_task_parameters(only_output_errors=True)

        if errors:
            failure = "Failed to save: parameter validation failed."
            if color:
                failure = colorize(failure, Colors.RED)
            print(failure)
            return self

        name = name or self.name
        user_blueprint_id = user_blueprint_id or self.user_blueprint_id
        # Update
        with self._workshop.auto_rate_limiter:
            if not save_as_new and user_blueprint_id is not None:
                user_blueprint = UserBlueprint.update(
                    self.serialize(),
                    user_blueprint_id=user_blueprint_id,
                    model_type=name,
                )
            else:
                # Insert
                user_blueprint = UserBlueprint.create(
                    self.serialize(),
                    model_type=name,
                    project_id=self._workshop.project_id,
                )

        return self._update_with_user_blueprint(
            user_blueprint, overwrite_blueprint_data=False
        )

    def delete(self, quiet=False):
        """
        Delete the blueprint from DataRobot

        Parameters
        ----------
        quiet: bool, Default=False
            If True, will not print a success message upon successful deletion.

        Raises
        ------
        ValueError
        """
        if not self.user_blueprint_id:
            raise ValueError(
                "Missing `user_blueprint_id`, has this BlueprintGraph been saved?"
            )
        with self._workshop.auto_rate_limiter:
            UserBlueprint.delete(user_blueprint_id=self.user_blueprint_id)
            if not quiet:
                print(colorize("Blueprint deleted.", Colors.SUCCESS))

    def link_to_project(
        self, project_id: Optional[str], user_blueprint_id: Optional[str] = None
    ):
        """
        Allows linking or unlinking of a user blueprint from a project.

        When a user blueprint is linked to a project, it will be able to validate and otherwise use
        project specific tasks, such as feature selection.

        If a user blueprint is linked to a project, it is expected that it will only be used with
        that project, or clones of that project.

        If a user blueprint is not linked to a project, it is expected that it should be able to be
        project-agnostic, that is, used with any project which shares the same target type as the
        user blueprint.

        Parameters
        ----------
        project_id: str or None
            If None is passed, unlink the user blueprint from the project
        user_blueprint_id: str or None
            If None is passed, fallback to `self.user_blueprint`, if also None, raise ValueError

        Raises
        ------
        ValueError

        """
        user_blueprint_id = user_blueprint_id or self.user_blueprint_id
        if not user_blueprint_id:
            raise ValueError(
                "Missing `user_blueprint_id`, has this BlueprintGraph been saved? "
                "If not, `user_blueprint_id` must be passed."
            )
        user_blueprint = UserBlueprint.update(
            None,
            user_blueprint_id=user_blueprint_id,
            project_id=project_id,
            include_project_id_if_none=True,
        )
        return self._update_with_user_blueprint(
            user_blueprint, overwrite_blueprint_data=False
        )

    def show(self, vertical=False, as_image=True) -> Optional[DisplayGraphviz]:
        """
        IPython specific drawing functionality.

        Wraps `as_image`.

        Parameters
        ----------
        vertical: bool (Default=False)
            Whether to display the graph top to bottom, rather than left to right
        as_image: bool (Default=True)
            If enabled, will render as a PNG, otherwise will rely on IPython to display
            the `Source` directly. Rendering as a PNG is preferred as it will auto-resize
            to the size of the rendering space provided, while `Source` will overflow.
        """
        return Visualize.show(
            self.as_graphviz_string(vertical=vertical), as_image=as_image
        )

    def as_graphviz_string(self, graph_name=None, vertical=False) -> str:
        """
        Used to render the graphviz form of the blueprint.

        Parameters
        ----------
        graph_name: string
            Name to display in the graph
        vertical: bool
            Whether to display the graph top to bottom, rather than left to right
        """
        if graph_name is None and self.name is not None:
            graph_name = self.name

        if self.creator is not None:
            graph_name += " \n(Created By: {})".format(self.creator)

        g = self._as_graph().copy()

        edges = []
        with self._frozen_tasks() as tasks:
            hash_mapping = {task.__hash__(): task.original_or_task_id for task in tasks}
            labels = {task.__hash__(): task.label for task in tasks}

        task_lookup = self._get_task_lookup()
        for task_id in g.nodes():
            if not isinstance(task_lookup.get(task_id), Task):
                _id = str(task_id)
                if represents_input_data(_id):
                    edges += [(Visualize.DATA_NODE, _id)]
                    labels[_id] = TaskInput.get_long_name(_id)
                else:
                    labels[_id] = _id

        labels[Visualize.DATA_NODE] = Visualize.DATA_NODE
        labels[Visualize.PREDICT_NODE] = Visualize.PREDICT_NODE
        for node in self._get_sink_task_ids(graph=g):
            g.add_edge(node, Visualize.PREDICT_NODE)

        g.add_node(Visualize.DATA_NODE)
        g.add_node(Visualize.PREDICT_NODE)

        for vw in g.edges():
            edge = []
            for e in vw:
                if isinstance(task_lookup.get(e), Task):
                    edge += [task_lookup.get(e).__hash__()]
                else:
                    edge += [e]
            edges += [edge]

        vertex_context_lookup = {
            item.task_id: item.messages for item in self.vertex_context
        }

        nodes_to_draw = [
            {"label": "{}".format(label), "id": node} for node, label in labels.items()
        ]

        return Visualize.to_graphviz_string(
            nodes_to_draw,
            edges,
            graph_name=graph_name,
            node_to_task_lookup=hash_mapping,
            vertex_context_lookup=vertex_context_lookup,
            vertical=vertical,
        )

    def to_source_code(
        self,
        to_stdout: bool = False,
        blueprint_ref: str = None,
        workshop_ref: str = None,
        pretty_repr: bool = True,
    ) -> str:
        """
        Convenience method for developers to help understand how to use the API.

        Construct the Blueprint using `deserialize`, and output the source code
        which can be used to create it using this new functional method.

        Parameters
        ----------
        to_stdout: bool (Default=False)
            Whether the source code of the blueprint should be printed to stdout.
        blueprint_ref: str (Default=None)
            A variable to store the new `BlueprintGraph` instead of generating one
        workshop_ref: str (Default=None)
            A variable to use to refer to an existing instance of `Workshop` instead
            of creating a new one.
        pretty_repr: bool (Default=True)
            Whether to pretty-print the result of execution. Useful for easy copy-paste
            in a notebook environment. Note: Does not change the content / actual value
            of what is returned, just how it is displayed (i.e. `repr` method).

        Returns
        -------
        str
            Executable source code to construct itself, in a functional manner.
            Useful as a tutorial, or large refactors.
        """
        variable_counter = {}
        variable_lookup = {}
        source_code_lines = []

        if self._user_blueprint:
            self._workshop.ensure_definitions_for_user_blueprint(self._user_blueprint)

        if workshop_ref is None:
            user_bp_project_id = (
                self._user_blueprint.project_id
                if self._user_blueprint is not None
                else None
            )
            project_id = user_bp_project_id or self._workshop.project_id
            workshop_args = dict()
            if self._user_blueprint:
                workshop_args[
                    "user_blueprint_id"
                ] = self._user_blueprint.user_blueprint_id

            if project_id is not None:
                workshop_args["project_id"] = project_id

            workshop_args_string = ", ".join(
                "{}={}".format(k, v.__repr__()) for k, v in workshop_args.items()
            )
            workshop_line = "w = Workshop({workshop_args_string})".format(
                workshop_args_string=workshop_args_string
            )
            source_code_lines += [workshop_line, ""]
            workshop_ref = "w"

        def build_variable(_task):
            """Build an appropriate variable name based on the task."""
            # Task code by default
            variable_name = _task.task_code.lower()

            # If we're selecting features, get the name if just one, otherwise
            # refer to it as 'selected_features'
            if len(_task.selected_features) > 0:
                variable_name = (
                    _task.selected_features[0].colname.lower()
                    if len(_task.selected_features) == 1
                    else "feature_selection"
                )

            # Retrieve count
            counter = variable_counter.get(variable_name, 0)

            # We added a variable, so increment the counter for next time
            variable_counter[variable_name] = counter + 1

            # If we've already got a variable of this name, add '_<count>' (starting at '_1')
            if counter > 0:
                variable_name += "_{}".format(counter)

            # Save it in our lookup table
            variable_lookup[_task.task_id] = variable_name

            return variable_name

        task = None
        variable_task_pairs = [
            (variable_lookup.get(task.task_id, build_variable(task)), task)
            for task in self._traverse_dag()
        ]

        # Make the proper functional calls for each task
        for variable, task in variable_task_pairs:
            # Construct the inputs for each functional call
            input_variables = ", ".join(
                [
                    self._built_input_ref(workshop_ref, variable_lookup, name)
                    for name in task.input_names
                ]
            )
            # Write the actual functional task call
            if task.selected_features:
                source_code_lines.extend(self._handle_feature_selection(task, variable))
            else:
                template = "{variable} = w.{select_task}({method_args})"
                method_args = "{inputs}".format(inputs=input_variables)
                if not task.using_default_output_method:
                    method_args += ", output_method={output_method}".format(
                        output_method="{w}.{o}".format(
                            w=workshop_ref,
                            o=TaskOutputMethod.get_variable_name(task.output_method),
                        )
                    )

                select_task = "Tasks.{task_code}".format(task_code=task.task_code)

                # Handle custom tasks
                if task.custom_task_id and task.version:
                    select_task = "CustomTasks.{task_code}_{task_id}".format(
                        task_code=task.task_code, task_id=task.custom_task_id
                    )
                    method_args += ", version={version}".format(
                        version=task.version.__repr__()
                    )

                source_code_lines.append(
                    template.format(
                        select_task=select_task,
                        variable=variable,
                        method_args=method_args,
                    )
                )

                parameter_assignments = ", ".join(
                    [
                        "{}={}".format(k, "'{}'".format(v) if isinstance(v, str) else v)
                        for k, v in task._minified_task_parameters.items()
                        if not (task.custom_task_id and k == "version_id")
                    ]
                )
                if parameter_assignments:
                    source_code_lines.append(
                        "{variable}.set_task_parameters({parameters})".format(
                            variable=variable, parameters=parameter_assignments
                        )
                    )

            # New line for readability
            source_code_lines.append("")

        if task is None:
            raise ValueError("Blueprint empty. Nothing to do.")

        # Save as the provided reference, if one was provided
        blueprint_name = (
            blueprint_ref
            if blueprint_ref
            else "{variable}_blueprint".format(variable=variable_lookup[task.task_id])
        )
        # Declare the final blueprint
        source_code_lines.append(
            "{blueprint_name} = {w}.BlueprintGraph({variable}, name={name})".format(
                w=workshop_ref,
                blueprint_name=blueprint_name,
                variable=variable_lookup[task.task_id],
                name=self.name if self.name is None else self.name.__repr__(),
            )
        )

        result = "\n".join(source_code_lines)
        if to_stdout:
            print(result)

        return PrettyString(result) if pretty_repr else result

    def _built_input_ref(self, workshop_ref, variable_lookup, name):
        return (
            "{w}.TaskInputs.{i}".format(w=workshop_ref, i=name)
            if represents_input_data(name)
            else variable_lookup[name]
        )

    def _handle_feature_selection(self, task, variable):
        source_code_lines = []
        if len(task.selected_features) == 1:
            template = "{variable} = w.Features.{feature_name}"
            source_code_lines.append(
                template.format(
                    variable=variable,
                    feature_name=task.selected_features[0].colname,
                )
            )
        else:
            by_type = {}
            for feature in task.selected_features:
                if feature.type in task.input_names:
                    by_type.setdefault(feature.type, []).append(feature.colname)

            for data_type, features in sorted(by_type.items(), key=lambda x: x[0]):
                template = (
                    "{variable} = w.FeatureSelection({features}, exclude={exclude})"
                )
                source_code_lines.append(
                    template.format(
                        variable=variable,
                        features=", ".join(["'{}'".format(f) for f in features]),
                        exclude=task.task_parameters.method == "exclude",
                    )
                )
        return source_code_lines

    def add_to_repository(
        self,
        project_id: Optional[str] = None,
        validate_parameters: bool = True,
    ) -> Optional[str]:
        """Add a user blueprint to a project's repository

        Parameters
        ----------
        project_id: str or None - required if Workshop not set to project
            the id of the project

        Returns
        -------
        str or None
            blueprint_id if the user blueprint was successfully
            added to the repository. None otherwise.
        validate_parameters: bool
            Ensure all parameters are valid (recommended)

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        project_id = project_id or self._workshop._project_id
        if project_id is None:
            raise NoProjectSpecifiedException(
                "The project id must be set in the Workshop or directly provided."
            )

        # If always save before adding to menu.
        self.save(validate_parameters=validate_parameters)

        with self._workshop.auto_rate_limiter:
            response = UserBlueprint.add_to_project(
                project_id=project_id, user_blueprint_ids=[self.user_blueprint_id]
            )
        added = response.added_to_menu
        if not added:
            if not self._user_blueprint or not self._user_blueprint.blueprint_id:
                print("Failure: Nothing was added to the project.")
                return None
            blueprint_id = self._user_blueprint.blueprint_id
        else:
            blueprint_id = added[0].blueprint_id
        return blueprint_id

    def train(
        self,
        project_id: Optional[str] = None,
        validate_parameters: bool = True,
        *args,
        **kwargs,
    ) -> BlueprintGraph:
        """
        Train it.

        See datarobot.Project.train or datarobot.Project.train_datetime
        for parameters.

        Parameters
        ----------
        project_id: str or None - required if Workshop not set to project
            the id of the project to train the blueprint on
        validate_parameters: bool
            Ensure all parameters are valid (recommended)

        Returns
        -------
        BlueprintGraph
            Returns self
        """
        project_id = project_id or self._workshop._project_id
        if project_id is None:
            raise NoProjectSpecifiedException(
                "The project id must be set in the Workshop or directly provided."
            )

        project = Project(project_id)
        bp = self.add_to_repository(project_id, validate_parameters=validate_parameters)
        if project.is_datetime_partitioned:
            project.train_datetime(bp, *args, **kwargs)
        else:
            project._train(bp, *args, **kwargs)
        print("Training requested! Blueprint Id: {}".format(bp))
        return self

    @classmethod
    def from_user_blueprint_tasks(
        cls,
        workshop: WorkshopInterface,
        user_blueprint_tasks: List[UserBlueprintTask],
        hex_column_name_lookup: Optional[
            List[UserBlueprintHexColumnNameLookupEntry]
        ] = None,
    ) -> BlueprintGraph:
        """Generate a BlueprintGraph from a list of UserBlueprintTask"""
        task_mapping = {
            user_blueprint_task.task_id: Task.from_user_blueprint_task(
                workshop=workshop,
                task_data=user_blueprint_task.task_data,
                original_id=user_blueprint_task.task_id,
                hex_column_name_lookup=hex_column_name_lookup,
            )
            for user_blueprint_task in user_blueprint_tasks
        }
        # Convert all inputs referencing ids into true references
        for task in task_mapping.values():
            task._map_inputs(task_mapping)

        return BlueprintGraph(
            workshop=workshop,
            final_task_or_tasks=list(task_mapping.values()),
        )

    @classmethod
    def from_user_blueprint(
        cls,
        workshop: WorkshopInterface,
        user_blueprint: UserBlueprint,
        creator: Optional[str] = None,
        last_modifier_name: Optional[str] = None,
    ) -> BlueprintGraph:
        """Create a BlueprintGraph from a UserBlueprint instance."""

        workshop.ensure_definitions_for_user_blueprint(user_blueprint)

        bp = cls.from_user_blueprint_tasks(
            workshop=workshop,
            user_blueprint_tasks=user_blueprint.blueprint,
            hex_column_name_lookup=user_blueprint.hex_column_name_lookup,
        )
        bp.name = user_blueprint.model_type
        bp.user_blueprint_id = user_blueprint.user_blueprint_id
        bp.project_id = user_blueprint.project_id
        bp.vertex_context = user_blueprint.vertex_context or []
        bp._user_blueprint = user_blueprint
        bp.creator = creator
        bp.last_modifier_name = last_modifier_name
        return bp

    @classmethod
    def deserialize(
        cls, workshop: WorkshopInterface, bp_json: List[Dict]
    ) -> BlueprintGraph:
        """Generate a BlueprintGraph from the JSON representation."""
        return BlueprintGraph.from_user_blueprint_tasks(
            workshop=workshop,
            user_blueprint_tasks=[
                UserBlueprintTask(**task_id_data_pair) for task_id_data_pair in bp_json
            ],
        )

    def serialize(self, minify=True) -> List[Dict]:
        """ "
        Used to view Blueprint in the familiar blueprint dictionary format.

        Returns
        -------
        list(dict)
            A list of dictionaries representing the tasks in the blueprint.
        """
        try:
            graph = self._as_graph(only_include_tasks=True)
            final_task: Task = self._get_task_lookup()[self._sorted_task_ids(graph)[-1]]
            self.tasks = list(set(self._tasks_from_task(final_task).values()))
            graph = self._as_graph(only_include_tasks=True)

            return [
                {
                    "task_id": self._get_task_lookup()[task_id].original_or_task_id,
                    "task_data": self._get_task_lookup()[task_id].serialize(
                        minify=minify
                    ),
                }
                for task_id in self._sorted_task_ids(graph)
                if task_id in self._get_task_lookup()
            ]
        except CycleException:
            raise ValueError("Blueprint is not a directed acyclic graph.")

    def share(self, usernames: Union[str, List[str]], role: str = Roles.CONSUMER):
        """Share a User Blueprint with a user or list of other users by username."""
        if self.user_blueprint_id is None:
            raise UnsavedBlueprintException(
                "Ensure the blueprint has been saved before sharing."
            )
        return self._workshop.share(
            self.user_blueprint_id, usernames=usernames, role=role
        )

    def share_with_group(
        self, group_ids: Union[str, List[str]], role: str = Roles.CONSUMER
    ):
        """Share a User Blueprint with a group or list of groups by id."""
        return self._workshop.share_with_group(
            user_blueprint_id=self.user_blueprint_id,
            group_ids=group_ids,
            role=role,
        )

    def share_with_org(
        self, org_ids: Union[str, List[str]], role: str = Roles.CONSUMER
    ):
        """Share a User Blueprint with a group or list of organizations by id."""
        return self._workshop.share_with_org(
            user_blueprint_id=self.user_blueprint_id,
            org_ids=org_ids,
            role=role,
        )

    # Methods below this point are not meant to be called directly by users.

    def __friendly_repr__(self) -> str:
        lookup = self._get_task_lookup()
        unique_task_names = set([])
        inputs = []
        task_names = []

        for task in self._traverse_dag():
            if task.label not in unique_task_names:
                task_names += [task.label]
            unique_task_names.add(task.label)
            for input_name in task.input_names:
                if input_name not in lookup:
                    input_label = self._input_name_lookup[input_name]
                    if input_label not in inputs:
                        inputs += [input_label]

        lines = [
            "Name: {}".format(
                "(Untitled)" if self.name is None else self.name.__repr__()
            ),
            "",
            "Input Data: " + " | ".join(inputs),
            "Tasks: " + " | ".join(task_names),
        ]

        if self.creator is not None:
            lines += ["Created By: {}".format(self.creator)]

        if self.last_modifier_name is not None:
            lines += ["Last Modified By: {}".format(self.last_modifier_name)]

        if self.project_id is not None:
            lines += ["Linked Project Id: {}".format(self.project_id)]

        return "\n".join(lines)

    def __eq__(self, other):
        if not isinstance(other, BlueprintGraph):
            return False
        return self.serialize() == other.serialize()

    @classmethod
    def _tasks_from_task(
        cls, task: Task, lookup: Optional[Dict[str, Task]] = None, freeze: bool = False
    ) -> Dict[str, Task]:
        """
        Build a list of tasks by finding and recursively
        finding all inputs to tasks.
        """
        lookup = lookup or {}

        task._freeze()

        lookup.update({task.task_id: task})
        inputs = [i for i in task.inputs if isinstance(i, Task)]
        for task in inputs:
            task._freeze()
            if isinstance(task, Task):
                lookup[task.task_id] = task
                # Is there an input not already stored in lookup?
                if any(
                    [
                        v.task_id not in lookup
                        for v in task.inputs
                        if isinstance(v, Task)
                    ]
                ):
                    # If so, make sure we pass the current lookup, and descend into
                    # the inputs which have not already been added.
                    # This can always be frozen as we will unfreeze at the outer-most call
                    lookup.update(cls._tasks_from_task(task, lookup, freeze=True))

        if not freeze:
            # Unfreeze all tasks to allow for modification
            for task in lookup.values():
                task._unfreeze()

        return lookup

    def _freeze(self):
        """
        Some fields such as the deterministic id of tasks and deterministic topological ordering
        require computation to find. If we can guarantee the tasks will not change, we can save
        significant computation, by using cached fields instead of recalculating them.
        """
        if self._frozen:
            return
        for task in self.tasks:
            if isinstance(task, Task):
                task._freeze()
        self._refresh_cached()
        self._frozen = True

    def _unfreeze(self):
        """
        Ready blueprint to be modified again.
        """
        for task in self.tasks:
            if isinstance(task, Task):
                task._unfreeze()
        self._frozen = False

    def _refresh_cached(self, attr_to_exclude=None):
        """Refresh cached attributes (set to None) of BlueprintGraph.

        Parameters
        ----------
        attr_to_exclude:
            A list of attributes which are not to be refreshed.
        """

        attr_to_refresh = ["_cached_graph"]
        refresh_cached(self, attr_to_refresh, attr_to_exclude)

    @contextmanager
    def _frozen_tasks(self, dag_order=False):
        """
        Freeze blueprint and iterate through the tasks.

        Parameters
        ----------
        dag_order: boolean
            Whether the tasks should be yielded in topologically sorted order.

        Yields
        ------
        list(Task)
        """
        frozen_before = self._frozen
        self._freeze()
        if dag_order:
            yield self._traverse_dag()
        else:
            yield self.tasks
        if not frozen_before:
            self._unfreeze()

    def _get_task_lookup(self) -> Dict[str, Task]:
        """
        Retrieve a lookup from hex representation to Task objects,
        or input data keys (e.g. 'NUM') to their identity.
        """
        with self._frozen_tasks() as tasks:
            return {hex_or_identity(v): v for v in tasks}

    def _is_dag(self, graph=None):
        """
        Determine whether the Blueprint is a DAG.

        Returns
        -------
        boolean
        """
        try:
            if graph is None:
                graph = self._as_graph(only_include_tasks=True)
            return nx.is_directed_acyclic_graph(graph)
        except CycleException:
            return False

    def _traverse_dag(self):
        """
        Simple, convenient, way to iterate through DAG in a topologically sorted manner.

        Note: Don't use this if modifying tasks as their hash will change during iteration
        and lookups will fail.

        Yields
        ------
        Task
        """
        not_dag = "Blueprint is not a directed acyclic graph, and cannot be reliably iterated."
        try:
            graph = self._as_graph(only_include_tasks=True)
            if not self._is_dag(graph):
                raise ValueError(not_dag)
            for task_id in self._sorted_task_ids(graph):
                yield self._get_task_lookup()[task_id]
        except CycleException:
            raise ValueError(not_dag)

    def _sorted_task_ids(self, graph=None):
        """
        Retrieve task ids in topologically sorted order.

        Estimators are preferred to be as late as possible, and if there are multiple possible
        topological sorts, this prioritizes sorts with estimators / models as late as possible.

        This is due to an expectation in DataRobot that models should follow tasks.

        Parameters
        ----------
        graph: DiGraph or None
            A graph to use instead of building it from scratch.

        Returns
        -------
        list(str)
        """
        # Ensure topological sort is deterministic.
        # If there is a tie, prefer estimators / models as late as possible,
        # otherwise sort lexicographically, and pick the first (arbitrary but deterministic)

        def get_sorting_name(task_hex):
            """
            0 comes before 1, so if it is an estimator, it will be chosen last to break a tie
            """
            is_estimator = self._get_task_lookup()[task_hex].is_estimator
            return "{}-{}".format(str(int(is_estimator)), task_hex)

        if graph is None:
            graph = self._as_graph(only_include_tasks=True)
        return list(nx.lexicographical_topological_sort(graph, key=get_sorting_name))

    def _as_graph(self, only_include_tasks=False):
        """
        Used to interpret the blueprint in the form of network

        Parameters
        ----------
        only_include_tasks: boolean. (Default=False)
            If false, will also add input data nodes, though they are not necessary for topological
            sorting (as they are always degree 0), and hurts performance.

        Returns
        -------
        DiGraph
            The Blueprint stored as a Directed Graph.
        """
        if self._frozen and self._cached_graph:
            return self._cached_graph

        g = nx.DiGraph()
        node_set = set()
        edge_set = set()

        def get_inputs(task):
            if only_include_tasks:
                return [
                    name for name in task.input_names if name in self._get_task_lookup()
                ]
            return task.input_names

        with self._frozen_tasks() as tasks:
            for task in tasks:
                inputs = get_inputs(task)
                task_hex = task.task_id
                node_set.add(task_hex)
                node_set |= {j for j in inputs}
                edge_set |= {(j, task_hex) for j in inputs}

        g.add_nodes_from(node_set)
        g.add_edges_from(edge_set)

        self._cached_graph = g
        return g

    def _update_with_user_blueprint(
        self, user_blueprint: UserBlueprint, overwrite_blueprint_data: bool = True
    ):
        """
        Parameters
        ----------
        user_blueprint: UserBlueprint

        overwrite_blueprint_data: bool

        Returns
        -------
        BlueprintGraph
            Returns self
        """
        self.name = user_blueprint.model_type
        self.project_id = user_blueprint.project_id
        self.user_blueprint_id = user_blueprint.user_blueprint_id
        self.vertex_context = user_blueprint.vertex_context or []
        self._user_blueprint = user_blueprint
        if overwrite_blueprint_data:
            temp_bp = self.from_user_blueprint_tasks(
                workshop=self._workshop,
                user_blueprint_tasks=user_blueprint.blueprint,
                hex_column_name_lookup=user_blueprint.hex_column_name_lookup,
            )
            self.tasks = temp_bp.tasks
        return self

    def _get_sink_task_ids(self, graph=None):
        """
        Retrieve all Sink task ids (no output edges).

        This can be useful for a number of applications including pruning, and
        finding the "predict" node.

        Parameters
        ----------
        graph: DiGraph (Default=None), if none provided, will create one.

        Returns
        -------
        list(str)
            A list of the nodes with no outgoing edges.
        """
        if graph is None:
            graph = self._as_graph().copy()
        return [node for node in graph.nodes() if len(graph.out_edges(node)) == 0]



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

import re
from typing import Type

from ._version import __version__
from .datarobot_client import UserBlueprint, APIObject
from .friendly_repr import FriendlyRepr
from .settings import Settings


class TaskInput(object):
    """
    The types available to use as input data in Tasks.
    """

    NUMERIC = "NUM"
    CATEGORICAL = "CAT"
    TEXT = "TXT"
    DATE = "DATE"
    DATE_DURATION = "DATE_DURATION"
    IMAGE = "IMG"
    COUNT_DICT = "COUNT_DICT"
    GEOSPATIAL = "GEO"
    ALL = "ALL"

    @classmethod
    def all(cls):
        return [
            cls.ALL,
            cls.CATEGORICAL,
            cls.COUNT_DICT,
            cls.DATE,
            cls.DATE_DURATION,
            cls.GEOSPATIAL,
            cls.IMAGE,
            cls.NUMERIC,
            cls.TEXT,
        ]

    @classmethod
    def get_long_name(cls, input_type):
        type_map = {
            cls.NUMERIC: "Numeric",
            cls.CATEGORICAL: "Categorical",
            cls.TEXT: "Text",
            cls.DATE: "Date",
            cls.DATE_DURATION: "Date Duration",
            cls.IMAGE: "Image",
            cls.COUNT_DICT: "Summarized Categorical",
            cls.GEOSPATIAL: "Geospatial",
            cls.ALL: "All",
        }
        return type_map[input_type]

    @classmethod
    def get_variable_name(cls, value):
        lookup = {
            cls.NUMERIC: "NUMERIC",
            cls.CATEGORICAL: "CATEGORICAL",
            cls.TEXT: "TEXT",
            cls.DATE: "DATE",
            cls.DATE_DURATION: "DATE_DURATION",
            cls.IMAGE: "IMAGE",
            cls.COUNT_DICT: "COUNT_DICT",
            cls.GEOSPATIAL: "GEOSPATIAL",
            cls.ALL: "All",
        }
        if value not in lookup:
            return '"ILLEGAL_INPUT"'
        return "TaskInput.{}".format(lookup[value])


ValidInputSet = set(TaskInput.all())


class TaskOutputMethod(object):
    """
    The methods which can be used with Tasks to produce output data.
    """

    TRANSFORM = "T"
    TRANSFORM_STACK = "TS"
    STACK_MARGIN = "Sm"
    STACK = "S"
    PREDICT = "P"
    PREDICT_MARGIN = "Pm"

    @classmethod
    def all(cls):
        return [
            cls.TRANSFORM,
            cls.TRANSFORM_STACK,
            cls.STACK,
            cls.STACK_MARGIN,
            cls.PREDICT,
            cls.PREDICT_MARGIN,
        ]

    @classmethod
    def is_estimator(cls, task_method):
        return task_method in (
            cls.STACK,
            cls.STACK_MARGIN,
            cls.PREDICT,
            cls.PREDICT_MARGIN,
        )

    @classmethod
    def get_variable_name(cls, value):
        lookup = {
            cls.TRANSFORM: "TRANSFORM",
            cls.TRANSFORM_STACK: "TRANSFORM_STACK",
            cls.STACK_MARGIN: "STACK_MARGIN",
            cls.STACK: "STACK",
            cls.PREDICT: "PREDICT",
            cls.PREDICT_MARGIN: "PREDICT_MARGIN",
        }
        if value not in lookup:
            return '"ILLEGAL_INPUT"'
        return "TaskOutputMethod.{}".format(lookup[value])


def represents_input_data(string: str) -> bool:
    """
    Check if a string represents input data (e.g. 'NUM' or 'AUDIO').

    Parameters
    ----------
    string: string
        The id of a task/task in the graph.
    """
    return string in ValidInputSet


def refresh_cached(obj, attr_to_refresh, attr_to_exclude=None):
    """Refresh cached attributes (set to None) of obj.

    Parameters
    ----------
    obj:
        The cached attributes of this object are to be refreshed.
    attr_to_refresh:
        A list of attributes to be refreshed.
    attr_to_exclude:
        A list of attributes which are not to be refreshed.
    """

    attr_to_refresh = set(attr_to_refresh)
    excluded = attr_to_exclude or []
    excluded = set(excluded)
    attr_to_refresh = attr_to_refresh - excluded
    for attr_name in attr_to_refresh:
        setattr(obj, attr_name, None)


def documentation(task_definition, auto_open=False):
    url_object = task_definition.url
    if url_object is not None:
        url_path = url_object.documentation
        if url_path is not None:
            url = UserBlueprint._client.domain + url_path
            if auto_open:
                open_link(url)
            return url
    print("No documentation available for this task.")


def open_link(url):
    """Try to directly open the provided url in the browser."""
    import sys

    if sys.platform == "win32":
        import os

        os.startfile(url)
        return

    import subprocess

    if sys.platform == "darwin":
        subprocess.Popen(["open", url])
        return

    try:
        subprocess.Popen(["xdg-open", url])
    except OSError:
        print("Please visit: {}".format(url))


def upper_case(match_obj):
    char_elem = match_obj.group(1)
    return char_elem.upper()


def safe_name(name):
    name = name.replace(" ", "_")
    name = re.sub(r"\W+", "", name)
    return re.sub(r"_(\w)", upper_case, name)


def task_ref(task):
    return "{}__{}".format(safe_name(task.label), task.task_code)


class PrettyString(FriendlyRepr, str):
    def __friendly_repr__(self):
        return self


class PrettyList(FriendlyRepr, list):
    def __friendly_repr__(self):
        return "\n\n".join([a.__repr__() for a in self])


class HiddenList(FriendlyRepr, list):
    def __friendly_repr__(self):
        return ""


def colorize(string: str, color: str):
    if not Settings().allow_color:
        return string

    """ Color a string for customized UX. """
    colors = {
        Colors.BLACK: "\u001b[30m",
        Colors.RED: "\u001b[31m",
        Colors.GREEN: "\u001b[32m",
        Colors.YELLOW: "\u001b[33m",
        Colors.BLUE: "\u001b[34m",
        Colors.MAGENTA: "\u001b[35m",
        Colors.CYAN: "\u001b[36m",
        Colors.WHITE: "\u001b[37m",
    }
    return "{}{}\x1b[0m".format(colors[color], string)


class Colors:
    BLACK = "black"
    RED = "red"
    GREEN = "green"
    YELLOW = "yellow"
    BLUE = "blue"
    MAGENTA = "magenta"
    CYAN = "cyan"
    WHITE = "white"
    # Meta-Colors
    SUCCESS = "green"
    ERROR = "red"
    WARNING = "yellow"


class Roles:
    OWNER = "OWNER"
    EDITOR = "USER"
    CONSUMER = "CONSUMER"


class RecipientType:
    USER = "user"
    GROUP = "group"
    ORGANIZATION = "organization"


def override_user_agent(api_object: Type[APIObject]) -> Type[APIObject]:
    """Override the user agent to show using blueprint workshop."""
    try:
        api_object._client.headers["User-Agent"] = "BlueprintWorkshop/{}".format(
            __version__
        )
    except Exception:
        print("Unable to override User-Agent, falling back to default")
    return api_object



"""
Copyright 2021 DataRobot, Inc. and its affiliates.

All rights reserved.

DataRobot, Inc. Confidential.

This is unpublished proprietary source code of DataRobot, Inc. and its affiliates.

The copyright notice above does not evidence any actual or intended publication of such source code.

Released under the terms of DataRobot Tool and Utility Agreement.
"""

import os
import tempfile

from IPython.core import magic_arguments
from IPython.core.magic import register_cell_magic
from datarobot import CustomModelVersion
from .datarobot_client import CustomTask


@magic_arguments.magic_arguments()
@magic_arguments.argument(
    "custom_task_id", type=str, help="Environment to use with the custom task"
)
@register_cell_magic
def update_custom(line, cell):
    args = magic_arguments.parse_argstring(update_custom, line)
    return update_custom_task(code=cell, custom_task_id=args.custom_task_id)


def update_custom_task(code: str, custom_task_id: str = None):
    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "custom.py"), mode="w") as f:
            f.write(code)

        custom_task = CustomTask.get(custom_task_id)
        previous_version = custom_task.latest_version
        environment_id = previous_version.base_environment_id

        return CustomModelVersion.create_from_previous(
            custom_task_id, environment_id, is_major_update=False, files=[f.name]
        ).custom_task_id
